# -*- coding: utf-8 -*-
# Copyright (C) 2014-2024  EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See https://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author : Konstantin Leontev (OpenCascade S.A.S)

import inspect
from traceback import format_exc

from qtsalome import QApplication, Qt, QTimer
from .geomrepairadv_logger import logger
from .geomrepairadv_worker_messages import MessageHandlerFactory
from multiprocessing import Process, Pipe

import CORBA


class Worker(Process):
    """
    Creates a process to run a given target function with a progress dialog as a parent.
    """

    def __init__(self, parent=None, target=None, args=None):
        super().__init__()

        # Set target function and its arguments
        self.target = target
        self.args = args

        # Create a pipe for communication
        self.parent_conn, self.child_conn = Pipe()

        # Set a timer to handle signals non-blocking way
        self.signals_timer = QTimer()
        self.signals_timer.timeout.connect(self.handle_signals)
        self.signals_timer.start(1000)

        # Create a message handler factory
        self.message_handler_factory = MessageHandlerFactory()

        # Store parent for signal handling
        self.parent = parent

        # Calculate total amount of lines in executed function
        source_lines = inspect.getsourcelines(target)
        total_lines = len(source_lines[0])
        first_line = source_lines[1]

        # Set a progress emitter to update the progress from the target
        self.progress_emitter = ProgressEmitter(self.child_conn, total_lines, first_line)

        # Set a variable for result
        self.result = None

        # Create a timer to terminate the thread if we can't exit gracefully
        self.terminate_timer = QTimer()
        self.terminate_timer.timeout.connect(self.terminate)
        self.terminate_timer.setSingleShot(True)


    def run(self):
        """
        Runs the given target function.
        """

        try:
            self.result = self.target(self.args, self.progress_emitter)

            # Reset the progress when finished
            self.child_conn.send(('progress', 100))

            if not self.is_canceled():
                self.child_conn.send(('completed', self.result))

        except Exception:
            logger.error(format_exc())
            self.child_conn.send(('failed',))


    def is_canceled(self):
        """
        Returns true if the job was canceled.
        """

        return not self.progress_emitter.is_running


    def cancel(self):
        """
        Cancels the job.
        """

        # Set emitter to know that we canceled the job and
        # can return from the running function gracefully.
        self.progress_emitter.stop()

        # Set a timer for a case when the running function
        # doesn't use emitter or does so heavy job that it would take
        # a lot of time for waiting and we need to terminate it.
        self.terminate_timer.start(1000)


    def terminate(self):
        """
        Overrides default terminate() to add some clean up.
        """

        # Check if the job is still running and need to be terminated
        if self.is_alive():
            logger.warning('Process will be terminated!')

            super().terminate()

        self.on_finished()


    def on_finished(self):
        """
        Called when the execution is finished.
        """

        logger.debug('on_finished called')

        QApplication.restoreOverrideCursor()
        self.signals_timer.stop()

        self.join()


    def get_result(self):
        """
        Returns result of the execution or None if the execution failed.
        """

        return self.result


    def start(self):
        """
        Starts the process and sets up the communication with the parent.
        """
        super().start()
        self.handle_signals()

        QApplication.setOverrideCursor(Qt.WaitCursor)


    def handle_signals(self):
        """
        Handles signals from the child process.
        """

        logger.debug('handle_signals called')

        if not self.parent_conn.poll():
            return

        msg = self.parent_conn.recv()
        logger.debug('msg: %s', msg)

        # Get a handler for the message type
        handler = self.message_handler_factory.get_handler(msg[0])
        if handler:
            handler.handle(msg, self)
        else:
            logger.error('No handler for message type: %s', msg[0])


class ProgressEmitter():
    """
    Helper class to reduce code repetition while update progress
    from a function executed in a separated thread.
    """

    def __init__(self, progress_update, total_lines, first_line):
        self.progress_update = progress_update
        self.first_line = first_line

        self.progress_percent = total_lines / 100.0
        logger.debug('self.progress_percent: %f', self.progress_percent)

        # Flag to check if we need to return from a function running in the thread
        self.is_running = True


    def stop(self):
        """
        Call this methid from the Worker for early return
        from a function executed in the separated thread.
        """

        self.is_running = False


    def emit(self):
        """
        Call this methid in a target function to update a progress value
        based on a currently executed line number.
        Returns False if the job was stopped. Should be used in the algo scripts
        for early return from the running function and finish the thread. 
        """

        line = inspect.getframeinfo(inspect.stack()[1][0]).lineno
        logger.debug('line: %d', line)
        progress_value = (line - self.first_line) / self.progress_percent
        logger.debug('progress_value: %d', progress_value)

        self.progress_update.send(['progress', int(progress_value)])

        return self.is_running
