# -*- coding: utf-8 -*-
# Copyright (C) 2014-2024  EDF
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
# See https://www.salome-platform.org/ or email : webmaster.salome@opencascade.com
#
# Author : Konstantin Leontev (OpenCascade S.A.S)

class MessageHandler:
    """
    MessageHandler is a base class for handling messages
    between child and worker processes in the progress bar Worker class.
    """

    def handle(self, msg, worker):
        """
        Handle the given message.

        This method should be implemented by subclasses to define how to process
        the given message.

        Args:
            msg: The message to be handled.
            worker: The worker context or object related to the message.

        Raises:
            NotImplementedError: If the method is not implemented by a subclass.
        """

        raise NotImplementedError("Subclasses should implement this method")


class ProgressHandler(MessageHandler):
    """
    ProgressHandler updates progress bar with a value from a message.
    """

    def handle(self, msg, worker):
        """
        Sets a new value for a worker progress bar.

        Args:
            msg (tuple): A tuple where the second element is the value to be set.
            worker (object): The worker object that has a setValue method.
        """

        worker.parent.setValue(msg[1])


class CompletedHandler(MessageHandler):
    """
    CompletedHandler handles the completion of the worker process.
    """

    def handle(self, msg, worker):
        """
        Handles the given message by performing actions on the worker object.

        Args:
            msg: The message to handle.
            worker: The worker object that contains the methods to be called.
        """

        worker.result = msg[1]

        worker.parent.on_completed()
        worker.on_finished()


class FailedHandler(MessageHandler):
    """
    FailedHandler handles the failure of the worker process.
    """

    def handle(self, msg, worker):
        """
        Handles the given message by invoking the worker's failure handler,
        stopping the timer, and joining the worker thread.

        Args:
            msg: The message to handle.
            worker: The worker object that contains the failure handler, timer, and worker thread.
        """

        worker.parent.on_failed()
        worker.on_finished()


class MessageHandlerFactory:
    """
    MessageHandlerFactory creates a handler for a given message type.
    """

    def __init__(self):
        self.handlers = {
            'progress': ProgressHandler(),
            'completed': CompletedHandler(),
            'failed': FailedHandler()
        }

    def get_handler(self, msg_type):
        """
        Returns a handler for the given message type.
        """

        return self.handlers.get(msg_type, None)
